  const { abs, floor } = Math;

export const UNITS = ['day', 'hour', 'minute', 'second'];
export const ISO_8601 = /^\d{4}(-\d\d(-\d\d(T\d\d:\d\d(:\d\d(,\d+)?)?(\.\d+)?(?<timezone>([+-]\d\d:?\d\d)|Z)?)?)?)?$/i;
const MS_IN_A_SECOND = 1000;
const MS_IN_A_MINUTE = 60 * MS_IN_A_SECOND;
const MS_IN_A_HOUR = 60 * MS_IN_A_MINUTE;
const MS_IN_A_DAY = 24 * MS_IN_A_HOUR;

const LABELS_DEFAULT = ['Days', 'Hours', 'Minutes', 'Seconds'];

export default class CountDownTimer extends HTMLElement {

  static get observedAttributes() {
    return ['alarm', 'date-timezone', 'date', 'labels'];
  }

  constructor() {
    super();
    this.state = {
      date: new Date().toISOString(),
      dateTimezone: 'fixed',
      alarm: '',
      labels: '',
      error: '',
    };
  }

  setState(newState) {
    this.state = {
      ...this.state,
      ...newState,
    };
    this.updateView();
  }

  connectedCallback() {
    const shadow = this.attachShadow({ mode: 'open' });
    this.element = document.createElement('div');
    this.element.setAttribute('part', 'timer-grid');
    this.setState({
      date: this.validateDate(this.getAttribute('date'), this.getAttribute('date-timezone')),
      dateTimezone: this.getAttribute('date-timezone'),
      alarm: this.getAttribute('alarm'),
      labels: this.getAttribute('labels'),
    });
    const test = new Date(null);
    const date = new Date(this.state.date);
    this.update(date - new Date());
    this.tick = setInterval(() => {
      const counter = date - new Date();
      this.update(counter);
    }, 1_000);
    const style = document.createElement('style');
    style.textContent =
      ':host{' +
        'display:inline-block;' +
      '}' +
      'div{' +
        'display:inline-grid;' +
        'justify-content:center;' +
        'grid-auto-flow:column;' +
        'gap:clamp(.5rem,.15em,1rem);' +
        'font-size:var(--cdt-digit-font-size,3rem);' +
        'line-height:1;' +
      '}' +
      'span{' +
        'font-size:1rem;' +
      '}' +
      '.pulse{' +
        'animation:ease-in-out pulse .25s alternate infinite;' +
      '}' +
      '@keyframes pulse{' +
        'from{' +
          'scale:.98;' +
        '}' +
        'to{' +
          'scale:1;' +
        '}' +
      '}'
    ;
    shadow.appendChild(style);
    shadow.appendChild(this.element);
  }

  updateView() {
    if (this.element) {
      if (this.state.error) {
        this.element.innerHTML = `<span>${this.state.error}</span>`;
      } else {
        const labels = LABELS_DEFAULT;

        if (typeof this.state.labels === 'string' && this.state.labels !== '') {
          this.state.labels.split('|', 4).forEach((label, index) => {
            const trimmed = label.trim();

            if (trimmed.length) {
              labels[index] = trimmed;
            }
          })
        }

        this.element.innerHTML = UNITS.map(
          (unit, index) => `<ct-number exportparts="number-grid, number-label, digit" label="${labels[index]}" unit="${unit}" number="0"></ct-number>`,
        ).join('');
      }
    }
  }

  /**
   * Validate and process the raw fate, if applicable.
   *
   * @param {string} dateIso8601 The raw date to validate.
   * @param {string} dateTimezone The method to handle relationship between date and timezone, either "fixed" or "flexible".
   *
   * @return {string|null} null will create a date for Jan, 01 1970 00:00:00.
   */
  validateDate(dateIso8601, dateTimezone = 'fixed') {
    if (dateIso8601 === undefined) return null;
    if (dateIso8601 === '') {
      this.setState({ error: 'No date provided.' });
      return null;
    }

    const match = dateIso8601.match(ISO_8601);
    if (!match) {
      this.setState({ error: 'No valid ISO 8601 date format provided.' });
      return null;
    }

    this.setState({ error: '' });

    // Real date and time differs from timezone, but the visible vales are
    // strictly the same.
    // For example, if the specified date is 2024-11-19T00:00:00+01:00, the
    // timezone part may change but not the date. This means that the counter
    // will stop everywhere around the world on November, 19 at 00:00:00.
    if (dateTimezone === 'flexible') {
      if (match.groups !== undefined && 'timezone' in match.groups && match.groups.timezone !== undefined) {
        dateIso8601 = dateIso8601.substring(0, dateIso8601.length - match.groups.timezone.length);
      }

      // Negate the offset value, as the difference is between the UTC date and
      // the local date, meaning that positive timezone leads to negative offset.
      const localTimezoneOffsetInHours = Math.floor(new Date().getTimezoneOffset() / -60);
      dateIso8601 = dateIso8601 + localTimezoneOffsetInHours.toLocaleString('en', {
        minimumIntegerDigits: 2,
        minimumFractionDigits: 2,
        maximumFractionDigits: 2,
        useGrouping: false,
        signDisplay: 'always'
      }).replace('.', ':');
    }

    return dateIso8601;
  }

  format(ms, unit) {
    let value;
    switch (unit) {
      case 'day':
        value = ms / MS_IN_A_DAY;
        break;
      case 'hour':
        value = (ms % MS_IN_A_DAY) / MS_IN_A_HOUR;
        break;
      case 'minute':
        value = (ms % MS_IN_A_HOUR) / MS_IN_A_MINUTE;
        break;
      case 'second':
        value = (ms % MS_IN_A_MINUTE) / MS_IN_A_SECOND;
        break;
      default:
        value = 0;
        break;
    }
    if (unit === 'day') return floor(value).toString();
    return floor(value).toString().padStart(2, '0');
  }

  update(counter) {
    if (counter <= 0) {
      clearInterval(this.tick);
      counter = 0;
    }
    UNITS.map((unit) => {
      const element = this.element.querySelector(`[unit="${unit}"]`);
      element?.setAttribute('number', this.format(abs(counter), unit));
    });
    const timeLeft = counter / MS_IN_A_SECOND;
    this.element.classList.toggle(
      'pulse',
      timeLeft < this.state.alarm && timeLeft > 0,
    );
  }

  attributeChangedCallback(name, oldValue, newValue) {
    if (name === 'date') {
      newValue = this.validateDate(newValue, this.state.dateTimezone);
    }
    else if (name === 'date-timezone') {
      const newDate = this.validateDate(this.state.date, newValue);
      this.setState({ date: newDate });
    }

    this.setState({ [name]: newValue });
  }
}
